# mypy: ignore-errors
# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off
# isort: skip_file
from .....schema_classes import IncidentAssigneeClass
from .....schema_classes import IncidentInfoClass
from .....schema_classes import IncidentSourceClass
from .....schema_classes import IncidentSourceTypeClass
from .....schema_classes import IncidentStageClass
from .....schema_classes import IncidentStateClass
from .....schema_classes import IncidentStatusClass
from .....schema_classes import IncidentTypeClass


IncidentAssignee = IncidentAssigneeClass
IncidentInfo = IncidentInfoClass
IncidentSource = IncidentSourceClass
IncidentSourceType = IncidentSourceTypeClass
IncidentStage = IncidentStageClass
IncidentState = IncidentStateClass
IncidentStatus = IncidentStatusClass
IncidentType = IncidentTypeClass

# fmt: on
